// proc.cpp for Linux  
//
// This program is free software. See the file COPYING for details.
// Author: Mattias Engdegrd, 1997-1999
// 		   Oliver


/* 
	LWP (Light Weight Process): just thread, mainly used in Solaris
	Task : thread in Linux
	NTPL(Native POSIX Thread Library)
*/  


#include <stdio.h>
#include <sys/types.h>
//#include <sys/param.h>	//HZ defined (no more used)
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sched.h>
#include <libgen.h> // basename()
#include <unistd.h> // sysconf()

#include <time.h>
#include "proc.h"
#include "svec.cpp"
#include "uidstr.h"
#include "ttystr.h"
#include "wchan.h"
#include "details.h"
#include "proc_common.cpp"

#include <QHash>

#define PROCDIR  "/proc"  // hmmm


float 	Procinfo::loadavg[] = {0.0, 0.0, 0.0};  // CPU load avg 1min,5min,15minute
float 	Procinfo::loadQps = 0.0;  
int 	Procinfo::qps_pid = -1;
int 	Procinfo::mem_total = 0;
int 	Procinfo::mem_free = 0;
int 	Procinfo::mem_buffers = 0;
int 	Procinfo::mem_cached = 0;
int 	Procinfo::swap_total = 0;
int 	Procinfo::swap_free = 0;
unsigned *Procinfo::cpu_times_vec = 0;
unsigned *Procinfo::old_cpu_times_vec = 0;
unsigned int Procinfo::boot_time = 0;
unsigned int Procinfo::clk_tick = 100; //for most system
bool Procinfo::socks_current = false;
QHash<int,Sockinfo*> Procinfo::socks;
QHash<int,UnixSocket*> Procinfo::usocks;
bool 	Procinfo::usocks_current = FALSE;
unsigned int Procinfo::num_cpus = 0;
unsigned int Procinfo::old_num_cpus = 0;
long 	Procinfo::num_process = 0;  //	64bit
long 	Procinfo::num_network_process = 0; // 64bit
long 	Procinfo::dt_system = 0; // diff system tick 

//DEL? int Procinfo::mem_shared = 0; // only linux kernel 2.4.x


// socket states, from <linux/net.h> and touched to avoid name collisions
enum {
	SSFREE = 0,			/* not allocated		*/
	SSUNCONNECTED,			/* unconnected to any socket	*/
	SSCONNECTING,			/* in process of connecting	*/
	SSCONNECTED,			/* connected to socket		*/
	SSDISCONNECTING			/* in process of disconnecting	*/
};

#define QPS_SCHED_AFFINITY ok

#ifdef QPS_SCHED_AFFINITY
#ifndef SYS_sched_setaffinity
#define SYS_sched_setaffinity 241
#endif
#ifndef SYS_sched_getaffinity
#define SYS_sched_getaffinity 242
#endif
// Needed for some glibc
int qps_sched_setaffinity(pid_t pid, unsigned int len, unsigned long *mask) {
	return syscall(SYS_sched_setaffinity, pid, len, mask);
};
int qps_sched_getaffinity(pid_t pid, unsigned int len, unsigned long *mask) {
	return syscall(SYS_sched_getaffinity, pid, len, mask);
};
#endif


/*
   When we commute from task group level to task level (or task level
   to task group level) we need to completley destroy the current state of
   proc[]. We exchange global <--> analytical informations. Some fields may
   not be accurate immediately : %WCPU for example.
   Some remarks : suppose we have a task group with TGID = 120, a thread leader 
   with PID=120 and two more threads with PID=125 and PID=127.
   In /proc, we only see /proc/120 and we don't see /proc/125 or /proc/127.
   This does not prevent the possibility to do an open("/proc/125", ...) or an
   open("/proc/127", ...). Also, path like /proc/120/task, /proc/125/task and
   /proc/127/task are ok.
   If the leader thread 120 terminates (it does un pthread_exit()), the directory
   /proc/120 will be maintained supposing 125 or 127 are alives. 
   Nevertheless, many files and directories will be empty in /proc/120. 
   For example, we lose "natual" access to threads 125 and 127 through /proc/120/task.
   At task group level, /proc/120/stat and /proc/120/status keep the global 
   informations ; so CPU time is still incresing. But, as task is empty, we see
   nothing for this task group at thread level.
   */

int Procinfo::page_k_shift;
int	pagesize;

struct proc_info_ {
	int  proc_id;
	char flag;
	char type;
	int  files;
} proc_info;

struct list_files_ {
	int proc_id;
	int flag;
	char *filename;	 //  path + filename 
} list_files;


// Bottleneck  
// DRAFT CODE 
// Description: read /proc/PID/fd/*  check opened file, count opened files
//		this fuction will be called  when every update.
// Return Value : 
int proc_pid_fd(const int pid)  
{
	char 	path[256];
	char 	buffer[256],fname[256];
	DIR 	*d;	
	int	    fdnum;
	int	    len, path_len;

	sprintf(path, "/proc/%d/fd", pid);

	path_len=strlen(path);
	d = opendir(path);

	if(!d){
		// this happend when the process died already or Zombie process
		// printf("Qps : read fail !! /proc/%d/fd !!! kernel bug ? \n",pid);
		return FALSE;
	}

	struct dirent *e;
	while((e = readdir(d)) != 0) 
	{
		if(e->d_name[0] == '.')
			continue;		// skip "." and ".."

		path[path_len]='/';
		path[path_len+1]=0;
		strcat(path,e->d_name);

		len = readlink(path, fname, sizeof(fname) - 1); 		
		if( len > 0) 
		{	
			fname[len]=0;
			//printf("DEBUG: %s[%s]\n",path,fname);
			//if (strcmp(fname,"/dev/null")==0 ) continue;	
		}

		num_opened_files++;
		//strcpy(p, e->d_name);
		//fdnum = atoi(p);
		//read_fd(fdnum, path);
	}
	closedir(d);
	return TRUE;	

}

// return the number of bytes read if ok, -1 if failed
int read_file_addNULL(char *name, char *buf, int max)
{
	int fd = open(name, O_RDONLY);
	if(fd < 0) return -1;
	int r = read(fd, buf, max);
	//buf[r]="\0";
	close(fd);
	return r;
}

int read_file(char *name, char *buf, int max)
{
	int fd = open(name, O_RDONLY);
	if(fd < 0) return -1;
	int r = read(fd, buf, max);
	close(fd);
	return r;
}


// new process created 
Procinfo::Procinfo(int pid) : refcnt(1)
{
	first_run=true;
	abs_row=-1;
	child_seq=-1;
	child_seq_prev=-1;
	
	lastchild=0;
	generation=-1;
	detail = 0;
	children = 0;
	fd_files = 0;
	maps = 0;

	socks_current = FALSE;
	usocks_current = FALSE;
	per_cpu_times = 0;

	size=0;
	resident=0;
	trs=0;
	drs=0;
	stack=0;
	share=0;

	environ = 0;
	envblock = 0;
	tgid=0;
	ppid=0;   // no parent
	
	pcpu=0;
	old_utime=0;
	selected = FALSE;
	hidekids = FALSE;

	Procinfo::pid=pid;

}

Procinfo::~Procinfo()
{
	if(detail) {
		detail->process_gone();
		detail = 0;
	}

	//    if(environ)    delete environ;
	if(envblock)
		free(envblock);


	if(maps) {
		maps->purge();
		delete maps;
	}
	if(fd_files) {
		fd_files->purge();
		delete fd_files;
	}

//	if(children)
//	{	children->clear();	delete children; }

	delete[] per_cpu_times;
}

// miscellaneous static initializations
void Procinfo::init_static()
{
	//socks.setAutoDelete(TRUE);
	///usocks.setAutoDelete(TRUE);

	page_k_shift = 0;
	for(int j = getpagesize(); j > 1024; j >>= 1) //4096
		page_k_shift++;

//	printf("pagesize=%d, %d\n ",getpagesize(), sysconf(_SC_PAGESIZE)); //4027
	pagesize=sysconf(_SC_PAGESIZE);
}



//  dirty_tricky function...(by fasthyun@magicn.com) 
//  Description : 
//		read /proc/1234/task/* and add to Proc::procs[]  
int Proc::read_pid_tasks(int pid)
{
	char 	path[256];
	struct 	dirent *e;
	int 	thread_pid;
	int 	thread_n=0;
	Procinfo *pi=0;

	sprintf(path, "/proc/%d/task", pid);

	DIR *d = opendir(path);
	if(!d) return -1;  // process dead  already!

	while((e = readdir(d)) != 0) {
		if(e->d_name[0] == '.') 	continue;	// skip "." , ".."

		thread_pid = atoi(e->d_name); 
		if (pid==thread_pid) continue; // skip
		
		pi=procs.value(thread_pid,NULL);

		if(pi==NULL)
		{
			pi = new Procinfo(thread_pid);
			procs.insert(pi->pid,pi);

		}
		if(pi->readproc(pid)>=0)
		{
			pi->generation = current_gen;
			if(pid!=thread_pid)
				pi->cmdline="(thread)";

		}
		thread_n++;
	}
	closedir(d);				
	return thread_n;
}

// update wcpu,%cpu field 
void Procinfo::calculate_cpu()  //
{
}


// Description :	read /proc/[PID]/*  or 
//					read /proc/[PID]/task/*
//      be called every UPDATE time. 
//      shoud be simple and clean !
//      return -1 means the process already dead !
int Procinfo::readproc(int tgid_)
{
	char path[128];     // enough
	char buf[512];     
	char sbuf[1024*2];  // should be enough to acommodate /proc/PID/stat
	char cmdbuf[MAX_CMD_LEN];
	int  cmdlen;
	int  len;
	char *p;

	if (flag_show_thread and flag_thread_ok and tgid_ > 0)  // a dirty_trick by fasthyun@magicn.com 
		sprintf(path, "/proc/%d/task/%d", tgid_,pid); 
	else
		sprintf(path, "/proc/%d", pid);

	//if (proc_pid_fd(pid)== TRUE) ;  // bottleneck !!

	if(first_run)  
	{
		old_wcpu=wcpu= pcpu = 0.0; // just a start
		
		//COMMAND , TGID, UID , COMMAND_LINE  never change ! 
		strcpy(buf, path);  
		strcat(buf, "/status");
		if((len = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) return -1;
		sbuf[len] = '\0';

		if(!(p = strstr(sbuf, "Name:")))
			return -1;
		sscanf(p, "Name: %s ",cmdbuf);
		command = cmdbuf;
		if(command=="qps") Procinfo::qps_pid=pid;
		if(command=="kthreadd") hidekids=true;
		if(command.contains("kthreadd")) hidekids=true;
		//printf("[%s]\n",command.toAscii().data()) ;

		if(!(p = strstr(sbuf, "Tgid:")))
			return -1;
		sscanf(p, "Tgid: %d ", &tgid);

	/*	if(!(p = strstr(sbuf, "PPid:")))
			return -1;
		sscanf(p, "PPid: %d ",&ppid); */
		
		if(!(p = strstr(sbuf, "Uid:")))
			return -1;
		sscanf(p, "Uid: %d %d %d %d Gid: %d %d %d %d",
				&uid, &euid, &suid, &fsuid,
				&gid, &egid, &sgid, &fsgid);

		username=userName(uid,euid);
		groupname=groupName(gid,egid);
		
		//COMMAND_LINE : read /proc/[PID]/cmdline
		strcpy(buf, path);
		strcat(buf, "/cmdline");
		if((cmdlen = read_file(buf, cmdbuf, MAX_CMD_LEN - 1)) < 0)
			return -1;

		// kernel 2.6.x bug :  a wrong comandline 
		int bug=0;
		if(cmdbuf[cmdlen-1]!=0) // how to check the bug
		{
			bug=true;
			cmdbuf[cmdlen] = 0;
		}

		if(cmdlen == 0) {
			cmdline = "";
		} else {
			//if(strlen(command.toAscii().data())==15)
			if(command.size()==15)
				if(!bug)
					command=basename(cmdbuf); // no memory leak !

			//change zero,0xA to ' ' 
			for(int i = 0; i < cmdlen-1; i++)
				if(cmdbuf[i]==0 or cmdbuf[i]==0x0A) cmdbuf[i] = ' ';
			
			// for non-ascii locale language 
			//cmdline = codec->toUnicode(cmdbuf,strlen(cmdbuf));
			cmdline = cmdbuf; 
			if(bug)
				//cmdline +=" >> Qps: may be a wrong commandline ";
				cmdline +=" ";

		}
		first_run=false;
	}

	// read /proc/1234/stat
	strcpy(buf, path);
	strcat(buf, "/stat");
	int statlen;
	if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) 
		return -1;
	sbuf[statlen] = '\0';

	// Not all values from /proc/#/stat are interesting; the ones left out
	// have been retained in comments to see where they should go, in case
	// they are needed again.
	//
	// In Linux 2.2.x, timeout has been removed, and the signal information
	// here is obsolete (/proc/#/status has real-time signal info).
	//
	// There are undocumented values after wchan, unused so far:
	// nswap		pages swapped out since process started
	// cnswap		nswap of children
	// exit_signal	(2.2.x) signal sent to parent when process exits
	// (The latter could provide a way to detect cloned processes since
	//  they usually have exit_signal != SIGCHLD, but I prefer waiting
	//  for real TIDs before implementing thread support.)
	//
	int  x_pid; // 
	long stime, cstime;
	int i_tty; //

	// Notes :
	//      1. ppid can be changed when parent dead !!
	sscanf(sbuf, "%d %s %c %d %d %d %d %d %lu %lu %lu %lu %lu "
			"%ld %ld %ld %ld %d %d %d %*s %lu %*s %*s %*s %*s %*s %*s %*s %*s "
			"%*s %*s %*s %*s %lu %*s %*s %*s %u %*s %*s %*s",&x_pid,&cmdbuf[0],
			&state, &ppid, &pgrp, &session, &i_tty, &tpgid,
			&flags, &minflt, &cminflt, &majflt, &cmajflt,
			&utime, &stime, &cutime, &cstime, &priority, &nice,&nthreads,
			/* itrealvalue */
			&starttime,
			/* vsize */
			/* rss */
			/* rlim, startcode, endcode, startstack kstkesp kstkeip,
			   signal, blocked, sigignore, sigcatch */
			&wchan,
			/* kernel 2.6 ? */
			/* 0L, 0L, exit_signal */
			&which_cpu
			/* rt_priority, policy */
	      );


	tty = (dev_t)i_tty;
	utime += stime;		// we make no user/system time distinction 
	cutime += cstime;
	
	//calculate_cpu();	// update wcpu,%cpu field 
	{
		int dcpu = utime - old_utime; // user_time from proc 
		if(dcpu<0 ) {
			    printf("Qps exception:[%d] dcpu=%d utime=%d pid_prev[%d]\n",pid,dcpu,utime,pid_prev);
			//    exit(0);
		}

		if(dt_system>0)	
			pcpu = num_cpus*100.0 * dcpu / dt_system;
		else // too fast read again ..
			printf("Qps exception: dt_system=%d  \n",dt_system);

		const float a = Procview::avg_factor;
		wcpu = a * old_wcpu + (1 - a) * pcpu;
	
		old_utime=utime;    // **** 
		old_wcpu=wcpu;
		old_which_cpu = which_cpu;
	
		if(dcpu==0 ) 
		{
			// I am a sleeping process 	
			// printf("[%d]dcpu zero\n",pid);
		}
	}    



	// read /proc/%PID/statm  - memory usage
	if(1) 
	{
		strcpy(buf, path);
		strcat(buf, "/statm");
		if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) 
			return -1;
		sbuf[statlen] = '\0';

		sscanf(sbuf, "%lu %lu %lu %lu %lu %lu %lu",
				&size, &resident, &share, &trs, &lrs, &drs, &dt);
		size	<<= page_k_shift;	// total memory
		resident<<= page_k_shift;	// 
		share	<<= page_k_shift;	// share*=pagesize;
		trs		<<= page_k_shift;	// text(code)
		lrs		<<= page_k_shift;	// zero : lib, awlays zero in Kernel 2.6
		drs		<<= page_k_shift;	// data: wrong in kernel 2.6
	//	dt		<<= page_k_shift;	// zero : in Kernel 2.6
		pmem = 100.0 * resident / mem_total;
	}

	// read /proc/1234/status check !!
	strcpy(buf, path);  
	strcat(buf, "/status");
	if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) return -1;
	sbuf[statlen] = '\0';
	int nfld;

	// slpavg : not supported in kernel 2.4; default value of -1
	if((p = strstr(sbuf, "SleepAVG:")))
		sscanf(p, "SleepAVG: %d",&slpavg);
	else 
		slpavg =-1;

	if(p = strstr(sbuf, "VmSize:"))
	{
		sscanf(p, "VmSize: %d",&size);
		//p = strstr(p, "VmRSS:");
		//if(p)
		//sscanf(p, "VmRSS: %d",&resident);
		if(p = strstr(p, "VmData:"))
			sscanf(p, "VmData: %d",&drs);			
		if(p = strstr(p, "VmStk:"))
			sscanf(p, "VmStk: %d",&stack);			
		if(p = strstr(p, "VmExe:"))
			sscanf(p, "VmExe: %d",&trs);
//		if(p = strstr(p, "VmLib:"))	sscanf(p, "VmLib: %d",&share);
	}

	/* 
	  generally Shared = RSS - ( CODE + DATA + STACK )
	   share= resident - trs -drs -stack;
	  
	 // Defines from task_mmu.c 
	 total_vm==size 
	 data = mm->total_vm - mm->shared_vm - mm->stack_vm;
	 swap = p->size - p->resident ; 	
	*/

	//which_cpu = 0;
	//per_cpu_times = 0;


	policy = -1;		// will get it when needed
	rtprio = -1;		// ditto
	tms = -1;		    // ditto
	
	
	pid_prev=pid; // possible ?

	modifed_bitarray=F_CPU;

	if(detail)	detail->set_procinfo(this);  // BAD !!!
	return pid;
}


// just grab the load averages
void Procinfo::read_loadavg()
{
	char path[80];
	char buf[512];
	int  n;
	strcpy(path, "/proc/loadavg");
	if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) {
		fprintf(stderr,	"qps: Cannot open /proc/loadavg  (make sure /proc is mounted)\n");
		exit(1);
	}
	buf[n] = '\0';
	sscanf(buf, "%f %f %f", &loadavg[0], &loadavg[1], &loadavg[2]);
}

// Description:  read common information  for all processes 
// return value 
// 		-1 : too fast refresh ! 
int Procinfo::read_common()
{
	static bool first_time=true;
	char path[80];
	char buf[1024*8]; // for SMP 

	char *p;
	int n;

	if(first_time)
	{
		/* Version 2.4.x ? */
		strcpy(path,"/proc/vmstat");
		if (!stat(path, (struct stat*)buf) )
			flag_24_ok=FALSE;
		else
			flag_24_ok=TRUE;

		/* NTPL(Native POSIX Thread Library) tasking ? */  
		strcpy(path,"/proc/1/task");
		if (!stat(path, (struct stat*)buf) )
			flag_thread_ok = TRUE;
		else
			flag_thread_ok = FALSE;

		strcpy(path, "/proc/stat"); 
		if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) return 0;
		buf[n] = '\0';
		p = strstr(buf, "btime");  
		if(p==NULL)
		{
			// not used ?  
			printf("Qps: A bug occurs ! [boot_time] \n"); 
			//boot_time= current time
		}else 
		{
			p+=6;
			sscanf(p, "%d", &boot_time);
		}
		first_time=false;
		clk_tick=sysconf(_SC_CLK_TCK);
		//printf("Qps: hz=%d\n",clk_tick); 
	}

	// read system status
	strcpy(path, "/proc/stat"); 
	if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) return 0;
	buf[n] = '\0';

	if(num_cpus==0) {	
		// count cpu for only static CPU system NOW !
		char *p;
		p = strstr(buf, "cpu");
		while(p < buf + sizeof(buf) - 4 && strncmp(p, "cpu", 3) == 0) 
		{
			num_cpus++;
			if(strncmp(p, "cpu0", 4) == 0)
				num_cpus--;
			p = strchr(p, '\n');
			if(p)
				p++;
		}
		cpu_times_vec = new unsigned[CPUTIMES * (num_cpus+2)];
		old_cpu_times_vec = new unsigned[CPUTIMES * (num_cpus+2)];

		//init
		for(unsigned cpu = 0; cpu < num_cpus + 1 ; cpu++)
			for(int i = 0; i < CPUTIMES; i++)
				CPU_TIMES(cpu, i)=0;
	}

	// DRAFT
	if(num_cpus != old_num_cpus) {
		// first time & changed cpu_num
		old_num_cpus=num_cpus;
	}


	// read total cpu  by dirty trick..
	int total_cpu=num_cpus;
	unsigned user,nice,system,idle,iowait, irq, sftirq, nflds;
	nflds = sscanf(buf, "cpu %u %u %u %u %u %u %u",
			&user, &nice,&system,&idle,&iowait, &irq, &sftirq);
	if( nflds > 4 ) {
		// kernel 2.6.x 
		system+=(irq+sftirq);
		idle+=iowait;
	}

	Procinfo::dt_system=idle - cpu_times(total_cpu, CPUTIME_IDLE)+
			     		user - cpu_times(total_cpu, CPUTIME_USER)+
				 		system - cpu_times(total_cpu, CPUTIME_SYSTEM);
	// printf("dt_system=%d\n",dt_system);	
	if(dt_system<clk_tick) // 1 sec less
	{
		///printf("\n==================== tooo fast =================================\n");
		//return -1; // too early refresh again  !!
	}
	else 

	// backup old values
	for(unsigned cpu = 0; cpu < num_cpus+1; cpu++)
		for(int i = 0; i < CPUTIMES; i++)
			old_cpu_times(cpu, i) = CPU_TIMES(cpu, i);
	
	cpu_times(total_cpu, CPUTIME_IDLE)=idle;
	cpu_times(total_cpu, CPUTIME_USER)=user;
	cpu_times(total_cpu, CPUTIME_SYSTEM)=system;
	
	{
	// SMP
	/*
		   in /proc/stat 
			cpu0	3350 	0	535	160879	1929	105	326	0
			cpu#,user,nice,system,idle,iowait,irq,sftirq
			
			Q1: kernel 2.4 cpu0 exist ?
	*/
		for(unsigned cpu = 0; cpu < num_cpus ; cpu++) {
			char cpu_buf[10];
			sprintf(cpu_buf, "cpu%d", cpu);
			if((p = strstr(buf, cpu_buf)) != 0) {
				nflds = sscanf(p, "%*s %u %u %u %u %u %u %u",
						&CPU_TIMES(cpu, CPUTIME_USER), &CPU_TIMES(cpu, CPUTIME_NICE),
						&CPU_TIMES(cpu, CPUTIME_SYSTEM), &CPU_TIMES(cpu, CPUTIME_IDLE),
						&iowait, &irq, &sftirq);

				if( nflds > 4 ) {
					// kernel 2.6.x 
					CPU_TIMES(cpu, CPUTIME_SYSTEM)+=(irq+sftirq);
					CPU_TIMES(cpu, CPUTIME_IDLE)+=iowait;
				}
				// 2.4.27-SMP bug 
				//if(old_cpu_times(cpu, i) >= CPU_TIMES(cpu, i) );	
				//if(old_cpu_times(cpu, CPUTIME_IDLE) >= CPU_TIMES(cpu, CPUTIME_IDLE) )
				//	CPU_TIMES(cpu, CPUTIME_IDLE)=old_cpu_times(cpu, CPUTIME_IDLE) ;

			} else {
				fprintf(stderr, "Error reading info for cpu%d (/proc/stat)\n", cpu);
				abort();
			}
		}
	}

	// read memory info
	strcpy(path, PROCDIR);
	strcat(path, "/meminfo");
	if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) return 0;
	buf[n] = '\0';

	// Skip the old /meminfo cruft, making this work in post-2.1.42 kernels
	// as well.  (values are now in kB)
	if( p = strstr(buf, "MemTotal:"))
		sscanf(p, "MemTotal: %d kB\n", &mem_total);
	if( (p = strstr(buf, "MemFree:")) != NULL )
		sscanf(p, "MemFree: %d kB\n", &mem_free);
	if( (p = strstr(buf, "Buffers:")) != NULL )
		sscanf(p, "Buffers: %d kB\n", &mem_buffers);
	if( (p = strstr(buf, "Cached:")) != NULL )
		sscanf(p, "Cached: %d kB\n", &mem_cached);

	p = strstr(buf, "SwapTotal:");
	sscanf(p, "SwapTotal: %d kB\nSwapFree: %d kB\n", &swap_total, &swap_free);

	return 0;	
}


int Procinfo::get_policy()
{
	if(policy == -1)
		policy = sched_getscheduler(pid);
	return policy;
}

int Procinfo::get_rtprio()
{
	if(rtprio == -1) {
		struct sched_param p;
		if(sched_getparam(pid, &p) == 0)
			rtprio = p.sched_priority;
	}
	return rtprio;
}
double Procinfo::get_tms()
{
	struct timespec ts;
	if(sched_rr_get_interval(pid, &ts)==-1)
		tms = -1; // should not be possible
	else {
		tms = ts.tv_nsec;
		tms /= 1000000000;
		tms +=ts.tv_sec;
	}
	return tms;
}

unsigned long Procinfo::get_affcpu()
{
#ifdef QPS_SCHED_AFFINITY
	if(qps_sched_getaffinity(pid, sizeof(unsigned long), &affcpu)==-1)
		affcpu=(unsigned long)0; 
#else
	if(sched_getaffinity(pid, sizeof(unsigned long), (cpu_set_t*)&affcpu)==-1)
		affcpu=(unsigned long)0; 
#endif
	return affcpu;
}

// Description : read  /proc/PID/fd/* (SYMBOLIC LINK NAME)  
/* We need to implement support for IPV6 and sctp ? */
void Procinfo::read_fd(int fdnum, char *path)
{
	int len;
	char buf[80];
	struct stat sb;

	// The fd mode is contained in the link permission bits
	if(lstat(path, &sb) < 0)
		return;
	int mode = 0;
	if(sb.st_mode & 0400) mode |= OPEN_READ;
	if(sb.st_mode & 0200) mode |= OPEN_WRITE;

	if( (len = readlink(path, buf, sizeof(buf) - 1)) > 0) {
		buf[len] = '\0';
		unsigned long dev, ino;
		if((buf[0] == '[' and sscanf(buf, "[%lx]:%lu", &dev, &ino) == 2	// Linux 2.0 style /proc/fd
				and dev == 0) 
				|| sscanf(buf, "socket:[%lu]", &ino) > 0) // Linux 2.1
		{
			//printf("sock ino=%d\n",ino);
			Sockinfo *si = NULL;
			si=Procinfo::socks.value(ino,NULL);
			char buf[80];
			if(si) {
				si->pid=pid;
				// a TCP or UDP socket
				sock_inodes.add(new SockInode(fdnum, ino));
				sprintf(buf, "%sp socket %lu",si->proto == Sockinfo::TCP ? "tc" : "ud", ino);
				fd_files->add(new Fileinfo(fdnum, buf, mode));
				return;
			} else {
				// maybe a unix domain socket?
				read_usockets();
				UnixSocket *us = NULL;
				if(usocks.contains(ino)) // for QHash
					us = usocks[ino];
				if(us) {
					char *tp = "?", *st = "?";
					switch(us->type) {
						case SOCK_STREAM: tp = "stream"; break;
						case SOCK_DGRAM: tp = "dgram"; break;
					}
					switch(us->state) {
						case SSFREE: st = "free"; break;
						case SSUNCONNECTED: st = "unconn"; break;
						case SSCONNECTING: st = "connecting"; break;
						case SSCONNECTED: st = "connected"; break;
						case SSDISCONNECTING: st = "disconn"; break;
					}
					sprintf(buf, "unix domain socket %lu (%s, %s) ",
							ino, tp, st);
					QString s = buf;
					s.append(us->name);
					fd_files->add(new Fileinfo(fdnum, s, mode));
					return;
				}
			}
		}
		// assume fds will be read in increasing order
		fd_files->add(new Fileinfo(fdnum, buf, mode));
	}
}


// Description : read /PID/fd opened files 
// return TRUE if /proc/PID/fd could be read, FALSE otherwise
// store fileinfo, and also socket inodes separately
// 
// called by Detail()
bool Procinfo::read_fds()
{
	char path[80], *p;
	if (flag_thread_ok && flag_show_thread)
		sprintf(path,"/proc/%d/task/%d/fd",pid,pid);
	else
		sprintf(path,"/proc/%d/fd", pid);

	DIR *d = opendir(path);
	if(!d) return false;

	if(!fd_files)
		fd_files = new Svec<Fileinfo*>(8);
	else 
		fd_files->purge();

	/*
	if(!sock_inodes) 	sock_inodes = new Svec<SockInode*>(4);
	else 	sock_inodes->purge(); */

	p = path + strlen(path) + 1;	// suck solution (by fasthyun@magicn.com) 
	p[-1] = '/';

	struct dirent *e;
	while((e = readdir(d)) != 0) {
		if(e->d_name[0] == '.')
			continue;		// skip . and ..
		strcpy(p, e->d_name);
		int fdnum = atoi(p);
		read_fd(fdnum, path);
	}
	closedir(d);
	return TRUE;
}

// tcp, udp
bool Procinfo::read_socket_list(Sockinfo::proto_t proto, char *pseudofile)
{
	char path[80];
	strcpy(path, PROCDIR);
	strcat(path, "/net/");
	strcat(path, pseudofile);
	FILE *f = fopen(path, "r");
	if(!f) return false;

	char buf[256];
	// Header 
	// sl  local_addr rem_addr  st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
  
	fgets(buf, sizeof(buf), f);	// skip header
	while(fgets(buf, sizeof(buf), f) != 0) {
		Sockinfo *si = new Sockinfo;
		si->pid=-1;
		si->proto = proto;
		unsigned local_port, rem_port, st, tr;
		sscanf(buf + 6, "%x:%x %x:%x %x %x:%x %x:%x %x %d %d %d",
				&si->local_addr, &local_port, &si->rem_addr, &rem_port,
				&st, &si->tx_queue, &si->rx_queue,
				&tr, &si->tm_when, &si->rexmits,
				&si->uid, &si->timeout, &si->inode);
		// fix fields that aren't sizeof(int)
		si->local_port = local_port;
		si->rem_port = rem_port;
		si->st = st;
		si->tr = tr;
		//printf("inode =%d \n",si->inode);
		socks.insert(si->inode, si);
		//linear_socks.add(si);
	}
	fclose(f);
	return true;	
}


bool Procinfo::read_usocket_list()
{
	char path[80];
	strcpy(path, PROCDIR);
	strcat(path, "/net/unix");
	FILE *f = fopen(path, "r");
	if(!f) return FALSE;

	char buf[256];
	fgets(buf, sizeof(buf), f);	// skip header
	while(fgets(buf, sizeof(buf), f)) {
		if(buf[0])
			buf[strlen(buf) - 1] = '\0'; // chomp newline
		UnixSocket *us = new UnixSocket;
		unsigned q;
		unsigned type, state;
		int n;
		sscanf(buf, "%x: %x %x %x %x %x %ld %n",
				&q, &q, &q, &us->flags, &type, &state, &us->inode, &n);
		us->name = buf + n;
		us->type = type;
		us->state = state;
		usocks.insert(us->inode, us);
	}
	fclose(f);
	return TRUE;
}

void Procinfo::read_sockets()
{
	//if(socks_current)	return;
	//socks.clear();
	if(!read_socket_list(Sockinfo::TCP, "tcp") || !read_socket_list(Sockinfo::UDP, "udp"))
		return;
	// remove no more socket ino
	socks_current = true;
}

void Procinfo::read_usockets()
{
	if(usocks_current)
		return;

	usocks.clear();
	if(!read_usocket_list())
		return;

	usocks_current = TRUE;
}

void Procinfo::invalidate_sockets()
{
	socks_current = usocks_current = false;
}

// return TRUE if /proc/XX/maps could be read, FALSE otherwise
bool Procinfo::read_maps()
{
	// idea: here we could readlink /proc/XX/exe to identify the executable
	// when running 2.0.x
	char name[80];

	if (flag_thread_ok && flag_show_thread)
		sprintf(name, "/proc/%d/task/%d/maps", pid,pid);
	else
		sprintf(name, "/proc/%d/maps", pid);

	FILE *f = fopen(name, "r");
	if(!f) return false;

	char line[1024];		// lines can be this long, or longer
	if(!maps)
		maps = new Svec<Mapsinfo *>;
	else
		maps->purge();

	while(fgets(line, sizeof(line), f)) {
		Mapsinfo *mi = new Mapsinfo;
		int n;
		sscanf(line, "%lx-%lx %4c %lx %x:%x %lu%n",
				&mi->from, &mi->to, mi->perm, &mi->offset,
				&mi->major, &mi->minor, &mi->inode, &n);
		if(line[n] != '\n') {
			int len = strlen(line);
			if(line[len - 1] == '\n')
				line[len - 1] = '\0';
			while(line[n] == ' ' && line[n]) n++;
			mi->filename = line + n;
		} else if((mi->major | mi->minor | mi->inode) == 0)
			mi->filename = "(anonymous)";
		maps->add(mi);
	}
	fclose(f);
	return TRUE;
}

// DRAFT CODE:
// return TRUE if /proc/1234/environ could be read, FALSE otherwise
bool Procinfo::read_environ()
{
	int     bsize=0;
	int		size;
	char    path[256];

	if(flag_thread_ok && flag_show_thread)
		sprintf(path, "/proc/%d/task/%d/environ", pid,pid);
	else
		sprintf(path, "/proc/%d/environ", pid);

	bsize=fsize(path);
	if(bsize<=0)
		return false;

	if(envblock)     free(envblock); // refresh()

	envblock = (char *)malloc(bsize+1);
	size=read_file(path,envblock,bsize+1);
	if(size<=0)
		return false;

	// kernel 2.6.x has a bug 
	if(envblock[size-2]==0) // how to check the bug.
	{
		char *p="Kernel Bug?=Qps: wrong environments !  please,check /proc/pid/environ !!";
		size=strlen(p)+1;
		if(bsize>size)
			strcpy(envblock,p);
	}

	int i=0,n=0,v=0;
	char ch;
	environ.purge();

	for(i=0; i<size ; i++)
	{
		ch=envblock[i];
		if(ch=='=')
		{
			envblock[i]=0;
			v=i+1;
		}
		if(ch==0)
		{
			environ.add(new NameValue(&envblock[n],&envblock[v]));
			//printf("%s %s\n",&envblock[n],&envblock[v]); 
			n=i+1;
		}
	}
	return true;
}


// ??
Cat_dir::Cat_dir(const char *heading, const char *explain, const char *dirname,
		QString Procinfo::*member)
: Cat_string(heading, explain),
	dir(dirname),
	cache(member)
{}

QString Cat_dir::string(Procinfo *p)
{
	if((p->*cache).isNull()) {
		char path[128], buf[512];

		if (flag_thread_ok && flag_show_thread)
			sprintf(path, "/proc/%d/task/%d/%s", p->pid,p->pid,dir);
		else
			sprintf(path, "/proc/%d/%s", p->pid, dir);

		int n = readlink(path, buf, sizeof(buf) - 1);
		if(n < 0) {
			// Either a kernel process, or access denied.
			// A hyphen is visually least disturbing here.
			p->*cache = "-";
			return p->*cache;
		} else if(buf[0] != '[') {
			// linux >= 2.1.x: path name directly in link
			buf[n] = '\0';
			p->*cache = buf;
			return p->*cache;
		}

		// Either a Linux 2.0 link in [device]:inode form, or a Solaris link.
		// To resolve it, we just chdir() to it and see where we end up.
		// Perhaps we should change back later?
		if(chdir(path) < 0) {
			p->*cache = "-";    // Most likely access denied
		} else {
			// getcwd() is fairly expensive, but this is cached anyway
			if(!getcwd(buf, sizeof(buf))) {
				p->*cache = "(deleted)";
			} else 
				p->*cache = buf;
		}
	}
	return p->*cache;
}

	Cat_state::Cat_state(const char *heading, const char *explain)
: Category(heading, explain)
{}

QString Cat_state::string(Procinfo *p)
{
	QString s("   ");
	int ni = p->nice;

	s[0] = p->state;
	s[1] = (p->resident == 0 && p->state != 'Z') ? 'W' : ' ';
	s[2] = (ni > 0) ? 'N' : ((ni < 0) ? '<' : ' ');
	return s;
}

// Cat_policy LINUX
	Cat_policy::Cat_policy(const char *heading, const char *explain)
: Category(heading, explain)
{}

QString Cat_policy::string(Procinfo *p)
{
	QString s;
	switch(p->get_policy()) {
		case SCHED_FIFO:
			s = "FI"; break;	// first in, first out
		case SCHED_RR:
			s = "RR"; break;	// round-robin
		case SCHED_OTHER:
			s = "TS"; break;	// time-sharing
		default:
			s = "??"; break;
	}
	return s;
}

int Cat_policy::compare(Procinfo *a, Procinfo *b)
{
	return b->get_policy() - a->get_policy();
}

	Cat_rtprio::Cat_rtprio(const char *heading, const char *explain)
: Category(heading, explain)
{}

QString Cat_rtprio::string(Procinfo *p)
{
	QString s;
	s.setNum(p->get_rtprio());
	return s;
}

int Cat_rtprio::compare(Procinfo *a, Procinfo *b)
{
	return b->get_rtprio() - a->get_rtprio();
}

	Cat_tms::Cat_tms(const char *heading,const char *explain)
: Category(heading,explain)
{}

QString Cat_tms::string(Procinfo *p)
{
	QString s;
	p->tms = p->get_tms();
	s.sprintf("%.3f",p->tms);
	return s;
}
int Cat_tms::compare(Procinfo *a, Procinfo *b)
{
	return (int)((b->get_tms() - a->get_tms())*1000) ;
}

	Cat_affcpu::Cat_affcpu(const char *heading,const char *explain)
: Category(heading,explain)
{}

QString Cat_affcpu::string(Procinfo *p)
{
	QString s;
	p->affcpu = p->get_affcpu();
	s.sprintf("%lx",p->affcpu);
	return s;
}
/*
   int Cat_affcpu::compare(Procinfo *a, Procinfo *b)
   {
   return (int)(b->affcpu - a->affcpu);
   }
   */

// LINUX?
	Cat_time::Cat_time(const char *heading, const char *explain)
: Category(heading, explain)
{}

QString Cat_time::string(Procinfo *p)
{
	QString s;
	char buff[64];
	int ticks = p->utime;
	int hundreds;	
	if(Qps::cumulative)
		ticks += p->cutime;

	int t = ticks / Procinfo::clk_tick;		// seconds

	if(t < 10) {
		hundreds = ticks / (Procinfo::clk_tick / 100) % 100;
		sprintf(buff,"%1d.%02ds", t, hundreds);
	} else if(t < 100 * 60) {
		sprintf(buff,"%2d:%02d", t / 60, t % 60);
	} else if(t < 100 * 3600) {
		int h = t / 3600; // 1hour = 3600 = 60m*60s
		t %= 3600;
		sprintf(buff,"%2d:%02dh", h, t / 60);
	} else {
		int d = t / 86400; // 1 day = 24* 3600s 
		t %= 86400;
		sprintf(buff,"%dd%dh", d, t / 3600);
	}
	s=buff;
	return s;
}

int Cat_time::compare(Procinfo *a, Procinfo *b)
{
	int at = a->utime, bt = b->utime;
	if(Qps::cumulative) {
		at += a->cutime;
		bt += b->cutime;
	}
	return bt - at;
}

	Cat_start::Cat_start(const char *heading, const char *explain)
: Category(heading, explain)
{}

QString Cat_start::string(Procinfo *p)
{
	time_t start = p->boot_time + p->starttime / Procinfo::clk_tick;
	QString s;
	char *ct = ctime(&start);
	time_t  tx= p->starttime / Procinfo::clk_tick;
	//if(p->tv.tv_sec - start < 86400) {
	if( tx < 86400) {
		ct[16] = '\0';
		s = ct + 11;
	} else 
	{
		ct[10] = '\0';
		s = ct + 4;
	}
	return s;
}

int Cat_start::compare(Procinfo *a, Procinfo *b)
{
	unsigned long bs = b->starttime, as = a->starttime;
	return bs >= as ? (bs == as ? 0 : 1) : -1;
}

Cat_percent::Cat_percent(const char *heading, const char *explain, int w,
	float Procinfo::*member)
: Category(heading, explain), float_member(member), field_width(w)
{}

QString Cat_percent::string(Procinfo *p)
{
	QString s;
	s.sprintf("%01.2f", (double)(p->*float_member));
	return s;
}

int Cat_percent::compare(Procinfo *a, Procinfo *b)
{
	float at = a->*float_member, bt = b->*float_member;
	return at < bt ? 1 : (at > bt ? -1 : 0);
}

// LINUX ?
Cat_tty::Cat_tty(const char *heading, const char *explain)
: Cat_string(heading, explain)
{}

QString Cat_tty::string(Procinfo *p)
{
	return Ttystr::name(p->tty);
}

Proc::Proc()
{
	// Note: When adding/removing/changing the fields, the save file
	// version must be increased!
	categories.insert(F_PID,  new Cat_int("PID", "Process ID", 6, &Procinfo::pid));
	categories.insert(F_TGID, new Cat_int("TGID","Task group ID ( parent of threads )",6, &Procinfo::tgid));
	categories.insert(F_PPID, new Cat_int("PPID", "Parent process ID", 6,	&Procinfo::ppid));
	categories.insert(F_PGID, new Cat_int("PGID", "Process group ID", 6,	&Procinfo::pgrp));
	categories.insert(F_SID,  new Cat_int("SID", "Session ID", 6,	&Procinfo::session));
	categories.insert(F_TTY,  new Cat_tty("TTY", "Terminal"));
	categories.insert(F_TPGID,new Cat_int("TPGID", "Process group ID of tty owner", 6, &Procinfo::tpgid));

	categories.insert(F_USER, new Cat_string("USER", "Owner (*=suid root, +=suid a user)",&Procinfo::username));
	categories.insert(F_GROUP,new Cat_string("GROUP", "Group name (*=sgid other)",&Procinfo::groupname));

	categories.insert(F_UID,  new Cat_int("UID", "Real user ID", 6, &Procinfo::uid));
	categories.insert(F_EUID, new Cat_int("EUID", "Effective user ID", 6,	&Procinfo::euid));
	categories.insert(F_SUID, new Cat_int("SUID", "Saved user ID (Posix)", 6,&Procinfo::suid));
	categories.insert(F_FSUID,new Cat_int("FSUID", "File system user ID", 6,&Procinfo::fsuid));
	categories.insert(F_GID,  new Cat_int("GID", "Real group ID", 6, &Procinfo::gid));
	categories.insert(F_EGID, new Cat_int("EGID", "Effective group ID", 6,&Procinfo::egid));
	categories.insert(F_SGID, new Cat_int("SGID", "Saved group ID (Posix)", 6,&Procinfo::sgid));
	categories.insert(F_FSGID,new Cat_int("FSGID", "File system group ID", 6,&Procinfo::fsgid));
	categories.insert(F_PRI,  new Cat_int("PRI", "Dynamic priority", 4,&Procinfo::priority));
	categories.insert(F_NICE, new Cat_int("NICE",	"Scheduling favour (higher -> less cpu time)",4, &Procinfo::nice));
	categories.insert(F_NLWP, new Cat_int("NLWP", "Number of tasks(threads) in task group",5, &Procinfo::nthreads));

	categories.insert(F_PLCY, new Cat_policy("PLCY", "Scheduling policy"));
	categories.insert(F_RPRI, new Cat_rtprio("RPRI","Realtime priority (0-99, more is better)"));
	categories.insert(F_TMS,  new Cat_tms("TMS", "Time slice in milliseconds"));
	categories.insert(F_SLPAVG, new Cat_int("%SAVG", "Percentage average sleep time (-1 -> N/A)",4,&Procinfo::slpavg));
	categories.insert(F_AFFCPU, new Cat_affcpu("CPUSET", "Affinity CPU mask (0 -> API not supported)"));  // ??? 
	categories.insert(F_MAJFLT, new Cat_uintl("MAJFLT", "Number of major faults (disk access)", 8, &Procinfo::majflt));
	categories.insert(F_MINFLT, new Cat_uintl("MINFLT", "Number of minor faults (no disk access)", 8, &Procinfo::minflt));
	// Memory
	categories.insert(F_SIZE, new Cat_memory("SIZE", "Virtual image size of process", 8, &Procinfo::size));
	categories.insert(F_RSS,  new Cat_memory("RSS", "Resident set size", 8, &Procinfo::resident));
	categories.insert(F_TRS,  new Cat_memory("TRS", "Text(code) resident set size", 8, &Procinfo::trs));
	categories.insert(F_DRS,  new Cat_memory("DRS", "Data resident set size(malloc+global variable)", 8, &Procinfo::drs));
	categories.insert(F_STACK,new Cat_memory("STACK", "Stack size", 8, &Procinfo::stack));
	categories.insert(F_SHARE,new Cat_memory("SHARE", "Shared memory with other libs", 8, &Procinfo::share));
	categories.insert(F_SWAP, new Cat_swap  ("SWAP", "Kbytes on swap device"));

	categories.insert(F_DT,   new Cat_uintl("DT", "Number of dirty (non-written) pages", 7, &Procinfo::dt));
	categories.insert(F_STAT, new Cat_state("STAT", "State of the process "));
	categories.insert(F_FLAGS,new Cat_hex("FLAGS", "Process flags (hex)", 9, &Procinfo::flags));
	categories.insert(F_WCHAN,new Cat_wchan("WCHAN", "Kernel function where process is sleeping"));
	categories.insert(F_WCPU, new Cat_percent("%WCPU", "Weighted percentage of CPU (30 s average)", 6, &Procinfo::wcpu));
	categories.insert(F_CPU,  new Cat_percent("%CPU", "Percentage of CPU used since last update", 6, &Procinfo::pcpu));
	categories.insert(F_MEM,  new Cat_percent("%MEM", "Percentage of memory used (RSS/total mem)", 6, &Procinfo::pmem));
	categories.insert(F_START,new Cat_start("START", "Time process started"));
	categories.insert(F_TIME, new Cat_time("TIME", "Total CPU time used since start"));
	categories.insert(F_CPUNUM,new Cat_int("CPU", "CPU the process is executing on", 3, &Procinfo::which_cpu));

	categories.insert(F_COMM, new Cat_string("COMMAND", "Command that started the process",	&Procinfo::command));
	categories.insert(F_CWD,  new Cat_dir("CWD", "Current working directory", "cwd", &Procinfo::cwd));
	categories.insert(F_ROOT, new Cat_dir("ROOT", "Root directory of process","root", &Procinfo::root));

	//command_line="COMMAND_LINE";	//reference to /proc/1234/cmdline
	categories.insert(F_CMDLINE, new Cat_cmdline("COMMAND_LINE", "Command line that started the process"));

	// java style
	QHashIterator<int,Category *> i(categories);  // a little suck!
	while (i.hasNext()) {
			i.next(); 
			i.value()->index=i.key();
			i.value()->id=i.key();
	}

	Procinfo::init_static();
	current_gen=0; // !
}


// Polling /proc/[PID]/*
void Proc::read_proc_all()
{
	DIR *d = opendir("/proc");
	struct dirent *e;

	while((e = readdir(d)) != 0) {
		if(e->d_name[0] >= '0' and e->d_name[0] <= '9') { // good idea !
			int pid;
			bool newbie=false;
			Procinfo *pi=NULL;

			pid=atoi(e->d_name);
//			if(pid<100) continue;
			
			pi=procs.value(pid,NULL);	
			
			if (pi==NULL)   // new process
			{
				pi = new Procinfo(pid);     
				procs.insert(pid,pi);
				newbie=true;
				pi->generation = current_gen; // ****
			}

			if(pi->readproc()>=0)
			{
				pi->generation = current_gen; // this process is alive

				/*
				   if(newbie){ 
				   Procinfo *parent;
				   parent =procs[pi->ppid];
				   if(parent)
				   {
				   parent->children->add(pi);        
				   }
				   else
				   {
				//root_procs.add(p);
				printf("Qps : parent null  pid[%d]\n",pi->pid);
				}        
				} */

				if(flag_show_thread and flag_thread_ok ) 
				{
					read_pid_tasks(pid);   //for threads 
				}

			}
			else 
				;// already /proc/PID  gone (dead) !!
		}        
	}
	closedir(d);
}

#include <sys/inotify.h> // glibc 
#include <errno.h> // glibc 

#define EVENT_SIZE		(sizeof (struct inotify_event))
#define BUF_LEN			(1024 * (EVENT_SIZE + 16))

#include <poll.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

// TEST 
void Proc::read_proc_inotify()
{
	char buf[BUF_LEN];
	int len, i = 0;
	if(fd==0) return;
	
	
	struct timeval time;
	fd_set rfds;
	int ret;

	/* timeout after five seconds */
	time.tv_sec = 0;
	time.tv_usec = 5;

	/* zero-out the fd_set */
	FD_ZERO (&rfds);

/*
 * add the inotify fd to the fd_set -- of course,
 * your application will probably want to add
 * other file descriptors here, too
 */
	FD_SET (fd, &rfds);
	ret = select (fd + 1, &rfds, NULL, NULL, &time);
	if (ret < 0)
        perror ("select");
	else if (!ret) return;
        /* timed out! */
	else if (FD_ISSET (fd, &rfds))
	{
		printf("xxxxxxxxxxxxxx EVEMT \n");
        /* inotify events are available! */
	len = read (fd, buf, BUF_LEN);
	printf(" len=%d\n",len);
	if (len < 0) {
        if (errno == EINTR)
                /* need to reissue system call */
                perror ("EINTR");
        else
                perror ("read");
	} else if (!len)
                perror ("small buffer!");

	while (i < len) 
	{
        struct inotify_event *event;
        event = (struct inotify_event *) &buf[i];

        printf ("wd=%d mask=%u cookie=%u len=%u",
                event->wd, event->mask,
                event->cookie, event->len);

        if (event->len)
                printf ("\tname=%s\n", event->name);
        else printf ("\n");

        i += EVENT_SIZE + event->len;
	}
	}

	return;
}

// Description: update the process list 
//		read /proc/*
// 		called by Procview::refresh(), every UPDATE .
void Proc::refresh()
{
	current_gen=!current_gen; // 0,1,0,1,0 ....

	static int init=0;
	static bool inotify_flag=false;
	if(false and init==0)
	{
		fd = inotify_init ();
		if (fd < 0)
			perror ("inotify_init");
		else 
			inotify_flag=true;

		int wd=0;
		wd = inotify_add_watch (fd, "/proc/3238",0x7FFFFFFF);//IN_ALL_EVENTS|IN_ISDIR|IN_UNMOUNT );// IN_MODIFY | IN_CREATE | IN_DELETE);
		
		//perror ("ixxxx");

		if (wd < 0)
			perror ("inotify_add_watch");
		init=1;
	}
	
	//init
	num_opened_files=0;
	Procinfo::num_process=0; 
	Procinfo::num_network_process=0; 
	if(Procinfo::read_common()<0) return; // **** should be updated every refresh !!
	Procinfo::read_sockets(); // for future

	if(inotify_flag)
	{
		read_proc_inotify();
	}
	else;
	
	read_proc_all();

}

// TEST
void Proc::refresh_inotify() 
{

}


// COMMON
Category *Proc::cat_by_name(const char *s)
{

	if(s) {
		// java style
		QHashIterator<int,Category *> i(categories);
		while (i.hasNext()) {
			const char *p;
			i.next();
			p=i.value()->name;
			if(*p==' ') p++;
			if(strcmp(p, s) == 0)
				return i.value();
//     cout << i.key() << ": " << i.value() << endl;
		}

	}
	return 0;
}

//COMMON: call by 
int  Proc::field_id_by_name(const char *s)
{
	if(s) {
		//STL style
		QHash<int,Category *>::iterator i = categories.begin();
		while (i != categories.end()) {
			if(strcmp(i.value()->name, s) == 0)	
				return i.key();	//cout << i.key() << ": " << i.value() << endl;
			++i;
		}	 	

	}	
	return -1;
}

// LINUX
int Procview::custom_fields[] = {F_PID, F_TTY, F_USER, F_NICE,
	F_SIZE, F_RSS,
	F_STAT, F_CPU, F_START, F_TIME,
	F_CMDLINE, F_END};

// basic field
int Procview::basic_fields[] = {F_PID, F_TTY, F_USER, F_NICE,
	F_STAT,F_RSS,
	F_MEM, F_CPU, F_START, F_TIME,
	F_CMDLINE, F_END};

int Procview::basic_fields_smp[] = {F_PID, F_TTY, F_USER,F_CPUNUM, F_NICE,
	F_STAT,F_RSS,
	F_MEM, F_CPU, F_START, F_TIME,
	F_CMDLINE, F_END};


int Procview::jobs_fields[] = {F_PID, F_TGID, F_PPID, F_PGID, F_SID, 
	F_TTY,F_TPGID, F_STAT, F_UID, F_TIME, F_CMDLINE, F_END};

int Procview::mem_fields[] = {
	F_PID, F_TTY, F_MAJFLT, F_MINFLT,
	F_SIZE,F_RSS,
	F_TRS, F_DRS,F_STACK,F_SHARE,
	//	F_DT,
	F_CMDLINE,F_END};

int Procview::sched_fields[] = {F_PID,F_TGID,F_NLWP,F_STAT,F_FLAGS,F_PLCY,
	F_PRI,F_NICE,F_TMS,F_SLPAVG,F_RPRI,F_AFFCPU,F_CPU,F_START,F_TIME,
	F_CMDLINE,F_END};


/*
   build_tree_24 is ok for 2.4 UP and SMP
   i don't know why  the new build_tree is ok for 2.4 UP, 2.6 UP and SMP
   but is not ok for 2.4 SMP
   */
void Procview::build_tree_24()
{ /*
	if(root_procs.size() > 0) {
		Procinfo *p;
		for(Q3IntDictIterator<Procinfo> it(proc->procs); (p = it.current()); ++it)
			if(p->children)
				p->children->clear();
		root_procs.clear();
	}
	Procinfo *p;
	for(Q3IntDictIterator<Procinfo> it(proc->procs); (p = it.current()); ++it) {
		if(accept_proc(p)) {
			Procinfo *parent = 0;
			if(p->ppid && (parent = proc->procs[p->ppid])
					&& accept_proc(parent)) {
				if(!parent->children)
					parent->children = new Svec<Procinfo *>(4);
				parent->children->add(p);
			} else
				root_procs.add(p);
		} else
			p->selected = FALSE;
	} */
}

//  
// 	called by Procview::rebuild()
// 	called every UPDATE,refresh
void Procview::build_tree()
{
	QHash<int,Procinfo*>::iterator i;
		
	Procinfo *p;
	if(root_procs.size() > 0) {  // children clear
		root_procs.clear();
	}
	// remove non-existing processes from procs[]
	for (i = procs.begin(); i != procs.end();)
	{
		Procinfo *p = i.value();
		//if(p==NULL)	printf("NulL\n");
		if(p->generation != current_gen) {
			//printf(" [ %s] \n",pi->command.ascii());
		 	delete p;
			i=procs.erase(i); 
			//procs.remove(pi->pid); 
		} 
		else
		{	
			p->children.clear();
			if(accept_proc(p)) 
				p->accepted=true;
			else
				p->accepted=false; 	
			++i;
		}
	}
	Procinfo::num_process=procs.size();      // count process	
	
	// find parent of a process
  	for (i = procs.begin(); i != procs.end();++i)
	{
		p=i.value();  // always not NULL?
		
		 /* p->children.clear();
		if(accept_proc(p)) 
			p->accepted=true;
		else
			p->accepted=false; 
		*/
		if( p->accepted) 
		{
			Procinfo *parent = 0;
			int ppid;	
			if(p->tgid==p->pid) 
				ppid=p->ppid;  // the process has no thread!!
			else
				//if(p->tgid!=p->ppid) printf("pid[%d]: tgid =%d , ppid %d \n",p->pid,p->tgid,p->ppid);
				ppid=p->tgid;
			
			if(p->pid < ppid)
			{	
				// this will occur !!	someone sended a report mail to me 
				//printf("ERROR : pid=%d ppid=%d !!!!!!!!!!!!!!!!!!!!!\n",p->pid,ppid);
				//exit(0);
			}
			parent = procs.value(ppid,NULL); 
			//if(parent==0) printf("thread_leader=0 (%d) %s\n",p->tgid,p->command.ascii());
			if(treeview and parent and parent->accepted )  
			{
	//			printf("(%d) %s\n",p->tgid,p->command.ascii());
				p->child_seq=parent->children.size();
				parent->children.add(p);
		
			} 
			else						
			{
				// 1.init(1) process  
				// 2.some process which parent not accepted 
				// 3.(null) thread has TGID=0,PPID=0
				// 4.linear mode
				root_procs.add(p);
				p->ppid=0;	// important !!!
			}

		} 
	}
	////printf("root_procs.size=%d\n",root_procs.size());
}

//BOTTLENECK No.3 
//Linux: re-sort the process list 
void Procview::rebuild()
{
	row_count=0;
	linear_procs.clear();  //Svec<Procinfo *> procs in Proview
	linear_socks.clear();
	
	Procinfo *pi = getProcinfoByPID(Procinfo::qps_pid);
	if(pi) Procinfo::loadQps=pi->pcpu;
	//else 

	build_tree();
	linearize_tree(&root_procs, 0, -1);
	
	/*
		// check process Acceptable
		for(Q3IntDictIterator<Procinfo> it(proc->procs); it.current(); ++it) {
			Procinfo *p = it.current();
			if(accept_proc(p))
				procs.add(p);
			else
				p->selected = FALSE;
		}
		static_sortcat = sortcat;
		procs.sort(reversed ? compare_backwards : compare);
	*/ 
}


// read new process info
void Procview::refresh()
{
	/****************************************************************/
	/* Procview.procs has the procinfo 								*/
	/****************************************************************/
	if(enable) 
	{	
		Proc::refresh(); 	// read "/proc/*", then update the process list
		rebuild();
	}
}

void Procview::set_fields()
{
	switch(viewfields) {
		case USER: // BASIC FIELD
			if(Procinfo::num_cpus>1)
				set_fields_list(basic_fields_smp);
			else
				set_fields_list(basic_fields);
			break;
		case JOBS:
			set_fields_list(jobs_fields);
			break;
		case MEM:
			set_fields_list(mem_fields);
			break;
		case SCHED:
			set_fields_list(sched_fields);
			break;
		case CUSTOM:
			set_fields_list(custom_fields);
			break;
		default:
			printf("Error ? set_fields_list \n");
	}
	fieldArrange();
}


// LINUX:
// deduce whether the currently selected fields correspond to a field list
void Procview::deduce_fields()
{
	return; // under development (by fasthyun@magicn.com) 2006/05/24

	if(viewfields != CUSTOM)
		return;

	Procview::fieldstates tags[4] = {USER, JOBS, MEM,SCHED};
	int *lists[4] = {basic_fields, jobs_fields, mem_fields,sched_fields};
	for(int i = 0; i < 4; i++) {
		int *l = lists[i];
		int j;
		for(j = 0; l[j] != F_END; j++)
			if(findCol(l[j]) < 0)
				break;
		if(l[j] == F_END && j == cats.size()) {
			viewfields = tags[i];
			return;
		}
	}
}


